# -*- coding: utf-8 -*-
"""
Aggregate powerrose and sort directions by dominance
"""
import h5py
import logging
import numpy as np
import pandas as pd

from reV.supply_curve.aggregation import Aggregation
from reV.supply_curve.points import SupplyCurveExtent

logger = logging.getLogger(__name__)


class WindDirections(Aggregation):
    """
    Aggregate PowerRose to Supply Curve points and sort directions in order
    of prominence. Then convert to equivalent sc_point_gid
    """
    DIR_ORDER = ['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW']

    def __init__(self, power_rose_h5_fpath, excl_fpath,
                 agg_dset='powerrose_100m', tm_dset='techmap_wtk',
                 resolution=64):
        """
        Parameters
        ----------
        power_rose_h5_fpath : str
            Filepath to .h5 file containing powerrose data
        excl_fpath : str
            Filepath to exclusions h5 with techmap dataset.
        agg_dset : str, optional
            Powerrose dataset to aggreate, by default 'powerrose_100m'
        tm_dset : str, optional
            Dataset name in the techmap file containing the
            exclusions-to-resource mapping data,
            by default 'techmap_wtk'
        resolution : int, optional
            SC resolution, must be input in combination with gid. Prefered
            option is to use the row/col slices to define the SC point instead,
            by default 64
        """
        super().__init__(excl_fpath, power_rose_h5_fpath, tm_dset, agg_dset,
                         resolution=resolution)

    @staticmethod
    def _map_direction_pos(power_rose_h5_fpath):
        """
        Map powerrose directions to sc row and column shifts

        Parameters
        ----------
        power_rose_h5_fpath : str
            Filepath to .h5 file containing powerrose

        Returns
        -------
        list
            Pos of major cardinal directions in power rose data
        """
        directions = WindDirections.DIR_ORDER

        with h5py.File(power_rose_h5_fpath, 'r') as f:
            cardinal_dirs = list(f['cardinal_directions'][...].astype(str))

        dir_pos = [cardinal_dirs.index(d) for d in directions]

        return dir_pos

    @staticmethod
    def _get_row_col_inds(sc_point_gids, row_length):
        """
        Convert supply curve point gids to row and col indices given row length

        Parameters
        ----------
        sc_point_gids : int | list | ndarray
            Supply curve point gid or list/array of gids
        row_length : int
            row length (shape[1])

        Returns
        -------
        row : int | list | ndarray
            row indices
        col : int | list | ndarray
            row indices
        """
        rows = sc_point_gids // row_length
        cols = sc_point_gids % row_length

        return rows, cols

    @staticmethod
    def _compute_neighbors(sc_point_gids, shape):
        """
        Compute neighboring supply curve point gids in following order:
        ['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW']

        Parameters
        ----------
        sc_point_gids : list | ndarray
            Supply curve point gids to get neighbors for
        shape : tuple
            Exclusions gid shape

        Returns
        -------
        neighbor_gids : ndarray
            Neighboring supply curve point gids
        """
        rows, cols = WindDirections._get_row_col_inds(sc_point_gids, shape[1])

        row_shifts = [-1, -1, 0, 1, 1, 1, 0, -1]
        rows = np.expand_dims(rows, axis=1) + row_shifts
        mask = rows < 0
        rows[mask] = 0
        mask = rows > shape[0]
        rows[mask] = shape[0]

        col_shifts = [0, 1, 1, 1, 0, -1, -1, -1]
        cols = np.expand_dims(cols, axis=1) + col_shifts
        mask = cols < 0
        cols[mask] = 0
        mask = cols > shape[1]
        cols[mask] = shape[1]

        neighbor_gids = rows * shape[1] + cols

        return neighbor_gids

    @staticmethod
    def _get_neighbors(excl_fpath, sc_point_gids, resolution=64):
        """
        Get neighboring sc_point_gids for all given supply curve points

        Parameters
        ----------
        excl_fpath : str
            Filepath to exclusions h5 with techmap dataset.
        sc_point_gids : list | ndarray
            List of supply curve point gids of interest
        resolution : int, optional
            SC resolution, must be input in combination with gid. Prefered
            option is to use the row/col slices to define the SC point instead,
            by default 64

        Returns
        -------
        neighbor_gids : pandas.DataFrame
            Neighboring sc_point_gids by cardinal direction
        """
        with SupplyCurveExtent(excl_fpath, resolution=resolution) as sc:
            shape = sc.shape

        neighbor_gids = \
            WindDirections._compute_neighbors(sc_point_gids, shape)

        directions = WindDirections.DIR_ORDER
        columns = ['{}_gid'.format(d) for d in directions]
        neighbor_gids = pd.DataFrame(neighbor_gids,
                                     columns=columns,
                                     index=sc_point_gids)

        return neighbor_gids

    def prominent_directions(self, excl_area=0.0081, max_workers=None,
                             chunk_point_len=100):
        """
        Aggregate power rose data to supply curve points, find all neighboring
        supply curve points, sort neighbors in order of prominent powerrose
        directions

        Parameters
        ----------
        excl_area : float, optional
            Area of an exclusion cell (square km), by default 0.0081
        max_workers : int | None, optional
            Number of cores to run summary on. None is all
            available cpus, by default None
        chunk_point_len : int, optional
            Number of SC points to process on a single parallel worker,
            by default 100

        Returns
        -------
        sc_pr : pandas.DataFrame
            Update meta data table with neighboring supply curve point gids
            and power-rose value at each cardinal direction
        """
        agg_out = self.aggregate(excl_area=excl_area, max_workers=max_workers,
                                 chunk_point_len=chunk_point_len)

        meta = agg_out.pop('meta')
        neighbor_gids = self._get_neighbors(self._excl_fpath,
                                            meta['sc_point_gid'].values,
                                            resolution=self._resolution)

        dir_pos = self._map_direction_pos(self._h5_fpath)
        sc_pr = agg_out.pop('powerrose_100m')[dir_pos].T

        columns = ['{}_pr'.format(d)
                   for d in WindDirections.DIR_ORDER]
        sc_pr = pd.DataFrame(sc_pr, index=meta['sc_point_gid'].values,
                             columns=columns)
        sc_pr = neighbor_gids.join(sc_pr)
        del neighbor_gids

        sc_pr.index.name = 'sc_point_gid'
        sc_pr = sc_pr.reset_index()
        sc_pr = pd.merge(meta, sc_pr, on='sc_point_gid')

        return sc_pr

    @classmethod
    def run(cls, power_rose_h5_fpath, excl_fpath,
            agg_dset='powerrose_100m', tm_dset='techmap_wtk',
            resolution=64, excl_area=0.0081, max_workers=None,
            chunk_point_len=100, out_fpath=None):
        """
        Aggregate powerrose to supply curve points, find neighboring supply
        curve point gids and rank them based on prominent powerrose direction

        Parameters
        ----------
        power_rose_h5_fpath : str
            Filepath to .h5 file containing power rose data, of same format as
            WTK data with directions on axis 0 (rows) and sites on axis 1
            (columns)
        excl_fpath : str
            Filepath to exclusions h5 with techmap dataset.
        agg_dset : str, optional
            Powerrose dataset to aggreate, by default 'powerrose_100m'
        tm_dset : str, optional
            Dataset name in the techmap file containing the
            exclusions-to-resource mapping data,
            by default 'techmap_wtk'
        resolution : int, optional
            SC resolution, must be input in combination with gid. Prefered
            option is to use the row/col slices to define the SC point instead,
            by default 64
        excl_area : float, optional
            Area of an exclusion cell (square km), by default 0.0081
        max_workers : int | None, optional
            Number of cores to run summary on. None is all
            available cpus, by default None
        chunk_point_len : int, optional
            Number of SC points to process on a single parallel worker,
            by default 100
        out_fpath : str
            Path to .csv file to save output table to

        Returns
        -------
        sc_pr : pandas.DataFrame
            Update meta data table with neighboring supply curve point gids
            and power-rose value at each cardinal direction
        """
        pr = cls(power_rose_h5_fpath, excl_fpath,
                 agg_dset=agg_dset, tm_dset=tm_dset,
                 resolution=resolution)

        sc_pr = pr.prominent_directions(excl_area=excl_area,
                                        max_workers=max_workers,
                                        chunk_point_len=chunk_point_len)

        if out_fpath:
            sc_pr.to_csv(out_fpath, index=False)

        return sc_pr



# -*- coding: utf-8 -*-
"""
Extract representative profiles for ReEDS
"""
import logging
import numpy as np
import pandas as pd
from reV.rep_profiles.rep_profiles import RepProfiles

from reVX.reeds.reeds_classification import ReedsClassifier
from reVX.utilities.exceptions import ReedsRuntimeError, ReedsValueError

logger = logging.getLogger(__name__)


class ReedsProfiles(RepProfiles):
    """
    Extract representative profile for ReEDS
    """
    def __init__(self, cf_profiles, rev_table, gid_col='gen_gids',
                 profiles_dset='cf_profile', rep_method='meanoid',
                 err_method='rmse', weight='gid_counts',
                 n_profiles=1, resource_classes=None,
                 region_map='reeds_region', cap_bins=5,
                 sort_bins_by='trans_cap_cost',
                 pre_filter=None, trg_by_region=False,
                 reg_cols=('region', 'class')):
        """
        Parameters
        ----------
        cf_profiles : str
            Filepath to reV gen output file to extract "cf_profile" from.
        rev_table : str | pd.DataFrame
            Aggregated supply curve summary table or file containing table
        gid_col : str
            Column label in rev_summary that contains the generation gids
            (data index in gen_fpath).
        profiles_dset : str
            Dataset name to pull generation profiles from.
        rep_method : str
            Method identifier for calculation of the representative profile.
        err_method : str
            Method identifier for calculation of error from the representative
            profile.
        weight : str | None
            Column in rev_summary used to apply weighted mean to profiles.
            The supply curve table data in the weight column should have
            weight values corresponding to the gid_col in the same row.
        n_profiles : int
            Number of representative profiles to save to fout.
        resource_classes : str | pandas.DataFrame | pandas.Series | dict
            Resource classes, either provided in a .csv, .json
            as a DataFrame or Series, or in a dictionary.
            If None, assumes rev_table has come from ReedsClassifier
        region_map : str | pandas.DataFrame
            Mapping of supply curve points to region to create classes for
        cap_bins : int
            Number of equal capacity bins to create for each
            region-class
        sort_bins_by : str | list, optional
            Column(s) to sort by before capacity binning,
            by default 'trans_cap_cost'
        pre_filter : dict | NoneType
            Column value pair(s) to filter on. If None don't filter
        trg_by_region : bool
            Groupby on region when computing TRGs
        reg_cols : tuple
            Label(s) for a categorical region column(s) to extract profiles
            for.
            Defaulted to ReedsClassifier region and class
        """
        if resource_classes is not None:
            rev_table = ReedsClassifier.create(rev_table, resource_classes,
                                               region_map=region_map,
                                               cap_bins=cap_bins,
                                               sort_bins_by=sort_bins_by,
                                               pre_filter=pre_filter,
                                               trg_by_region=trg_by_region)[0]

        super().__init__(cf_profiles, rev_table, reg_cols, gid_col=gid_col,
                         cf_dset=profiles_dset, rep_method=rep_method,
                         err_method=err_method, weight=weight,
                         n_profiles=n_profiles)

        if 'timezone' not in self.meta:
            msg = ('Meta data must contain timezone to enable '
                   'conversion to local time!')
            logger.error(msg)
            raise ReedsRuntimeError(msg)

    @staticmethod
    def _to_hourly(profiles, time_index):
        """
        Reduce profiles to hourly resolution

        Parameters
        ----------
        profiles : dict
            dict of n_profile-keyed arrays with shape (time, n) for the
            representative profiles for each region.
        time_index : pd.DatatimeIndex
            Datetime Index for represntative profiles

        Returns
        ----------
        profiles : dict
            dict of n_profile-keyed arrays with shape (time, n) for the
            representative profiles for each region at hourly resolutions
        time_index : pd.DatatimeIndex
            Datetime Index for represntative profiles reduced to hourly
            resolution
        """
        mask = time_index.minute == 0
        time_index = time_index[mask]
        profiles = {k: v[mask] for k, v in profiles.items()}

        return profiles, time_index

    @staticmethod
    def _to_hour_ending(profiles, time_index):
        """
        Shift profiles and time index to hour ending from instantaneous at
        top of hour. First timestep will be at 01:00 in current year, last
        timestep will be 00:00 in next year.

        Parameters
        ----------
        profiles : dict
            dict of n_profile-keyed arrays with shape (8760, n) for the
            representative profiles for each region.
        time_index : pd.DatatimeIndex
            Datetime Index for represntative profiles

        Returns
        ----------
        profiles : dict
            dict of n_profile-keyed arrays with shape (8760, n) for the
            representative profiles for each region, rolled to hour ending
        time_index : pd.DatatimeIndex
            Datetime Index for represntative profiles shifted to hour ending
        """
        if len(time_index) != 8760:
            e = ('ReedsProfiles cannot be shifted to hour ending unless they '
                 'are hourly! Got time index with length {}'
                 .format(len(time_index)))
            logger.error(e)
            raise ValueError(e)

        year = time_index.year[0]
        time_index = pd.date_range(start='{}0101'.format(year),
                                   end='{}0101'.format(int(year + 1)),
                                   freq='1h', closed='right')
        mask = (time_index.month == 2) & (time_index.day == 29)
        # drop 12/31 on leap years
        if any(mask):
            mask = (time_index.month == 12) & (time_index.day == 31)
            time_index = time_index[~mask]

        profiles = {k: np.roll(v, -1, axis=0) for k, v in profiles.items()}

        return profiles, time_index

    @staticmethod
    def _get_hour_of_year(time_index):
        """
        Compute the hour of the year from time_index

        Parameters
        ----------
        time_index : pandas.DatatimeIndex
            Datetime index to extract hours of the year from

        Returns
        -------
        hour_of_year : ndarray
            Vector of the hour of the year for each timestep
        """
        hour_of_year = (time_index.hour + (time_index.minute / 60)
                        + (time_index.dayofyear - 1) * 24)
        if np.max(hour_of_year) > len(time_index):
            hour_of_year = np.arange(len(time_index), dtype='int32')

        return hour_of_year

    @staticmethod
    def _roll_array(arr, shifts):
        """
        Roll array with unique shifts for each column
        This converts timeseries to local time
        Parameters
        ----------
        arr : ndarray
            Input timeseries array of form (time, sites)
        shifts : ndarray | list
            Vector of shifts from UTC to local time
        Returns
        -------
        local_arr : ndarray
            Array shifted to local time
        """
        if arr.shape[1] != len(shifts):
            msg = ('Number of timezone shifts ({}) does not match number of '
                   'sites ({})'.format(len(shifts), arr.shape[1]))
            logger.error(msg)
            raise ReedsValueError(msg)

        local_arr = np.zeros(arr.shape, dtype=arr.dtype)
        for i, s in enumerate(shifts):
            local_arr[:, i] = np.roll(arr[:, i], int(s))

        return local_arr

    def _to_local_time(self):
        """
        Shift profiles to local time from UTC
        """
        tz = self.meta['timezone'].values.copy()
        tz *= len(self.time_index) // 8760
        # pylint: disable=W0201
        self._profiles = {k: self._roll_array(v, tz)
                          for k, v in self.profiles.items()}

    def _run(self, fout=None, hourly=True, hour_ending=True, max_workers=None):
        """
        Extract ReEDS representative profiles in serial or parallel
        and save to disc if requested. Convert to hourly and hour_ending if
        requested.

        Parameters
        ----------
        fout : str, optional
            filepath to output h5 file, by default None
        hourly : bool, optional
            Flag to get hourly data (top of hour) instead of native resolution,
            by default True
        hour_ending : bool, optional
            Flag to shift instantaneous profiles and time index to hour ending,
            by default True
        max_workers : int, optional
            Number of parallel workers. 1 will run serial, None will use all
            available., by default None

        Returns
        -------
        profiles : dict
            dict of n_profile-keyed arrays with shape (time, n) for the
            representative profiles for each region.
        meta : pd.DataFrame
            Meta dataframes recording the regions and the selected rep profile
            gid.
        time_index : pd.DatatimeIndex
            Datetime Index for represntative profiles
        """
        if max_workers == 1:
            self._run_serial()
        else:
            self._run_parallel(max_workers=max_workers)

        self._to_local_time()

        # pylint: disable=W0201
        if hourly and (len(self._time_index) > 8760):
            self._profiles, self._time_index = self._to_hourly(self.profiles,
                                                               self.time_index)

        if hourly and hour_ending:
            self._profiles, self._time_index = \
                self._to_hour_ending(self.profiles, self.time_index)

        if fout is not None:
            self.save_profiles(fout, save_rev_summary=False,
                               scaled_precision=True)

        logger.info('Representative profiles complete!')

        return self.profiles, self.meta, self.time_index

    @classmethod
    def run(cls, cf_profiles, rev_table, gid_col='gen_gids',
            profiles_dset='cf_profile', rep_method='meanoid',
            err_method='rmse', weight='gid_counts',
            n_profiles=1, resource_classes=None, region_map='reeds_region',
            cap_bins=5, sort_bins_by='trans_cap_cost',
            pre_filter=None, trg_by_region=False,
            reg_cols=('region', 'class'), fout=None, hourly=True,
            hour_ending=True, max_workers=None):
        """Run representative profiles.

        Parameters
        ----------
        cf_profiles : str
            Filepath to reV gen output file to extract "cf_profile" from.
        rev_table : str | pd.DataFrame
            Aggregated supply curve summary table or file containing table
        gid_col : str
            Column label in rev_summary that contains the generation gids
            (data index in gen_fpath).
        profiles_dset : str
            Dataset name to pull generation profiles from.
        rep_method : str
            Method identifier for calculation of the representative profile.
        err_method : str
            Method identifier for calculation of error from the representative
            profile.
        weight : str | None
            Column in rev_summary used to apply weighted mean to profiles.
            The supply curve table data in the weight column should have
            weight values corresponding to the gid_col in the same row.
        n_profiles : int
            Number of representative profiles to save to fout.
        resource_classes : str | pandas.DataFrame | pandas.Series | dict
            Resource classes, either provided in a .csv, .json
            as a DataFrame or Series, or in a dictionary.
            If None, assumes rev_table has come from ReedsClassifier
        region_map : str | pandas.DataFrame
            Mapping of supply curve points to region to create classes for
         cap_bins : int
            Number of equal capacity bins to create for each
            region-class
        sort_bins_by : str | list, optional
            Column(s) to sort by before capacity binning,
            by default 'trans_cap_cost'
        pre_filter : dict | NoneType
            Column value pair(s) to filter on. If None don't filter
        trg_by_region : bool
            Groupby on region when computing TRGs
        reg_cols : tuple
            Label(s) for a categorical region column(s) to extract profiles
            for.
        fout : str, optional
            filepath to output h5 file, by default None
        hourly : bool, optional
            Flag to get hourly data (top of hour) instead of native resolution,
            by default True
        hour_ending : bool, optional
            Flag to shift instantaneous profiles and time index to hour ending,
            by default True
        max_workers : int, optional
            Number of parallel workers. 1 will run serial, None will use all
            available., by default None

        Returns
        -------
        profiles : dict
            dict of n_profile-keyed arrays with shape (time, n) for the
            representative profiles for each region.
        meta : pd.DataFrame
            Meta dataframes recording the regions and the selected rep profile
            gid.
        time_index : pd.DatatimeIndex
            Datetime Index for represntative profiles
        """
        rp = cls(cf_profiles, rev_table, gid_col=gid_col,
                 profiles_dset=profiles_dset, rep_method=rep_method,
                 err_method=err_method, weight=weight,
                 n_profiles=n_profiles, resource_classes=resource_classes,
                 region_map=region_map, cap_bins=cap_bins,
                 sort_bins_by=sort_bins_by, pre_filter=pre_filter,
                 trg_by_region=trg_by_region, reg_cols=reg_cols)

        rp._run(fout=fout, hourly=hourly, hour_ending=hour_ending,
                max_workers=max_workers)

        return rp.profiles, rp.meta, rp.time_index



# -*- coding: utf-8 -*-
"""
RPM Command Line Interface
"""
import json
import click
import logging
from reV.utilities.cli_dtypes import STR
from reV.utilities.loggers import init_mult

from reVX.rpm.rpm_manager import RPMClusterManager as rpm_cm
from reVX.rpm.rpm_output import RPMOutput as rpm_o

logger = logging.getLogger(__name__)


@click.group()
@click.option('--name', '-n', default='RPM', type=STR,
              help='Job name. Default is "RPM".')
@click.option('--cf_profiles', '-cf', required=True,
              type=click.Path(exists=True),
              help=('Path to reV .h5 file containing desired capacity factor '
                    'profiles'))
@click.option('--out_dir', '-o', required=True, type=click.Path(),
              help='Directory to dump output files')
@click.option('--parallel', '-p', is_flag=True,
              help='Run clustering in parallel by region. Default is serial.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, cf_profiles, out_dir, parallel, verbose):
    """
    RPM Command Line Interface
    """
    ctx.ensure_object(dict)
    ctx.obj['NAME'] = name
    ctx.obj['CF_FPATH'] = cf_profiles
    ctx.obj['OUT_DIR'] = out_dir
    ctx.obj['PARALLEL'] = parallel

    init_mult(name, out_dir, modules=[__name__, 'reVX.rpm'],
              verbose=verbose)

    logger.info('Running reV to RPM pipeline using CF profiles: {}'
                .format(cf_profiles))
    logger.info('Outputs to be stored in: {}'.format(out_dir))


@main.group(invoke_without_command=True)
@click.option('--rpm_meta', '-m', required=True, type=click.Path(exists=True),
              help='.csv or .json containing the RPM meta data')
@click.option('--region_col', '-rc', type=str, default=None,
              help='The meta-data field to map RPM regions to')
@click.pass_context
def cluster(ctx, rpm_meta, region_col):
    """
    Cluster RPM Regions
    """
    name = ctx.obj['NAME']
    cf_fpath = ctx.obj['CF_FPATH']
    out_dir = ctx.obj['OUT_DIR']
    parallel = ctx.obj['PARALLEL']
    ctx.obj['RPM_META'] = rpm_meta
    ctx.obj['META_COL'] = region_col

    if ctx.invoked_subcommand is None:
        logger.info('Clustering regions based on:\n{}'
                    .format(rpm_meta))
        rpm_cm.run_clusters(cf_fpath, rpm_meta, out_dir, job_tag=name,
                            rpm_region_col=region_col, parallel=parallel)


@cluster.command()
@click.option('--exclusions', '-e', default=None,
              type=click.Path(exists=True),
              help=('Filepath to exclusions data (must match the techmap grid)'
                    ' None will not apply exclusions.'))
@click.option('--excl_dict', '-exd', default=None, type=STR,
              help='String representation of a dictionary of exclusion '
              'LayerMask arguments {layer: {kwarg: value}} where layer is a '
              'dataset in excl_fpath and kwarg can be "inclusion_range", '
              '"exclude_values", "include_values", "use_as_weights", '
              'or "weight".')
@click.option('--techmap_dset', '-tmd', default=None, type=STR,
              help=('Dataset name in the exclusions file containing the '
                    'exclusions-to-resource mapping data.'))
@click.option('--trg', '-trg', default=None, type=STR,
              help=('Filepath to TRG LCOE bins.'))
@click.pass_context
def and_profiles(ctx, exclusions, excl_dict, techmap_dset, trg):
    """
    Cluster RPM Regions and extract representative profiles
    """
    name = ctx.obj['NAME']
    cf_fpath = ctx.obj['CF_FPATH']
    out_dir = ctx.obj['OUT_DIR']
    parallel = ctx.obj['PARALLEL']
    rpm_meta = ctx.obj['RPM_META']
    region_col = ctx.obj['META_COL']
    out_dir = ctx.obj['OUT_DIR']
    parallel = ctx.obj['PARALLEL']

    logger.info('Clustering regions based on: {}'.format(rpm_meta))
    logger.info('Extracting representative profiles using exclusions: {}'
                .format(exclusions))

    if isinstance(excl_dict, str):
        excl_dict = excl_dict.replace('\'', '\"')
        excl_dict = excl_dict.replace('None', 'null')
        excl_dict = excl_dict.replace('True', 'true')
        excl_dict = excl_dict.replace('False', 'false')
        excl_dict = json.loads(excl_dict)

    output_kwargs = {}
    if trg is not None:
        output_kwargs = {'trg': trg}
        logger.info('Applying TRGs: {}'.format(trg))

    rpm_cm.run_clusters_and_profiles(cf_fpath, rpm_meta, exclusions,
                                     excl_dict, techmap_dset, out_dir,
                                     job_tag=name, rpm_region_col=region_col,
                                     parallel=parallel,
                                     output_kwargs=output_kwargs)


@main.group(invoke_without_command=True)
@click.option('--rpm_clusters', '-rc', required=True,
              type=click.Path(exists=True),
              help=('Pre-existing RPM cluster results .csv with '
                    '(gid, gen_gid, cluster_id, rank)'))
@click.option('--exclusions', '-e', default=None,
              type=click.Path(exists=True),
              help=('Filepath to exclusions data (must match the techmap grid)'
                    ' None will not apply exclusions.'))
@click.option('--excl_dict', '-exd', default=None, type=STR,
              help='String representation of a dictionary of exclusion '
              'LayerMask arguments {layer: {kwarg: value}} where layer is a '
              'dataset in excl_fpath and kwarg can be "inclusion_range", '
              '"exclude_values", "include_values", "use_as_weights", '
              'or "weight".')
@click.option('--techmap_dset', '-tmd', default=None, type=STR,
              help=('Dataset name in the techmap file containing the '
                    'exclusions-to-resource mapping data.'))
@click.option('--trg', '-trg', default=None, type=STR,
              help=('Filepath to TRG LCOE bins.'))
@click.pass_context
def rep_profiles(ctx, rpm_clusters, exclusions, excl_dict, techmap_dset, trg):
    """
    Extract representative profiles from existing RPM clusters
    """
    name = ctx.obj['NAME']
    cf_fpath = ctx.obj['CF_FPATH']
    out_dir = ctx.obj['OUT_DIR']
    parallel = ctx.obj['PARALLEL']
    ctx.obj['RPM_CLUSTERS'] = rpm_clusters

    if ctx.invoked_subcommand is None:
        logger.info('Extracting representative profiles from RPM clusters: {}'
                    .format(rpm_clusters))
        logger.info('Extracting representative profiles using exclusions: {}'
                    .format(exclusions))

        if trg is not None:
            logger.info('Applying TRGs: {}'.format(trg))

        if isinstance(excl_dict, str):
            excl_dict = excl_dict.replace('\'', '\"')
            excl_dict = excl_dict.replace('None', 'null')
            excl_dict = json.loads(excl_dict)

        rpm_o.process_outputs(rpm_clusters, cf_fpath, exclusions,
                              excl_dict, techmap_dset, out_dir,
                              job_tag=name, parallel=parallel, trg=trg)


@rep_profiles.command()
@click.option('--profiles', '-np', required=True, type=int,
              help=('Number of profiles per cluster to export.'))
@click.option('--forecast_fpath', '-fo', type=STR, default=None,
              help=('reV generation output file for forecast data. If this is '
                    'input, profiles will be taken from forecast file instead '
                    'of the cf file, based on a NN mapping.'))
@click.pass_context
def extra_profiles(ctx, profiles, forecast_fpath):
    """
    Cluster RPM Regions and extract representative profiles
    """
    name = ctx.obj['NAME']
    cf_fpath = ctx.obj['CF_FPATH']
    out_dir = ctx.obj['OUT_DIR']
    parallel = ctx.obj['PARALLEL']
    rpm_clusters = ctx.obj['RPM_CLUSTERS']

    logger.info('Extracting extra representative profiles from: {}'
                .format(rpm_clusters))
    if forecast_fpath is not None:
        logger.info('Using forecast file: {}'.format(forecast_fpath))

    rpm_o.extract_profiles(rpm_clusters, cf_fpath, out_dir,
                           n_profiles=profiles, job_tag=name,
                           parallel=parallel, forecast_fpath=forecast_fpath)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running RPM CLI')
        raise



# -*- coding: utf-8 -*-
"""
ResourceX Command Line Interface
"""
import click
import logging
import os
import pandas as pd
from reV.utilities.loggers import init_mult

from reVX.resource.resource import ResourceX

logger = logging.getLogger(__name__)


@click.group()
@click.option('--resource_h5', '-h5', required=True,
              type=click.Path(exists=True),
              help=('Path to Resource .h5 file'))
@click.option('--out_dir', '-o', required=True, type=click.Path(),
              help='Directory to dump output files')
@click.option('--compute_tree', '-t', is_flag=True,
              help='Flag to force the computation of the cKDTree')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, resource_h5, out_dir, compute_tree, verbose):
    """
    ResourceX Command Line Interface
    """
    ctx.ensure_object(dict)
    ctx.obj['H5'] = resource_h5
    ctx.obj['OUT_DIR'] = out_dir
    ctx.obj['CLS'] = ResourceX
    ctx.obj['TREE'] = compute_tree

    name = os.path.splitext(os.path.basename(resource_h5))[0]
    init_mult(name, out_dir, verbose=verbose, node=True,
              modules=[__name__, 'reVX.resource.resource',
                       'reV.handlers.resource'])

    logger.info('Extracting Resource data from {}'.format(resource_h5))
    logger.info('Outputs to be stored in: {}'.format(out_dir))


@main.command()
@click.option('--lat_lon', '-ll', nargs=2, type=click.Tuple([float, float]),
              default=None,
              help='(lat, lon) coordinates of interest')
@click.option('--gid', '-g', type=int, default=None,
              help='Resource gid of interest')
@click.pass_context
def sam_file(ctx, lat_lon, gid):
    """
    Extract all datasets needed for SAM for the nearest pixel to the given
    (lat, lon) coordinates OR the given resource gid
    """
    if lat_lon is None and gid is None:
        click.echo("Must supply '--lat-lon' OR '--gid'!")
        raise click.Abort()
    elif lat_lon and gid:
        click.echo("You must only supply '--lat-lon' OR '--gid'!")
        raise click.Abort()

    with ctx.obj['CLS'](ctx.obj['H5'], compute_tree=ctx.obj['TREE']) as f:
        if lat_lon is not None:
            SAM_df = f.get_SAM_lat_lon(lat_lon)
        elif gid is not None:
            SAM_df = f.get_SAM_gid(lat_lon)

    out_path = "{}.csv".format(SAM_df.name)
    out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
    logger.info('Saving data to {}'.format(out_path))
    SAM_df.to_csv(out_path)


@main.command()
@click.option('--dataset', '-d', type=str, required=True,
              help='Dataset to extract')
@click.option('--lat_lon', '-ll', nargs=2, type=click.Tuple([float, float]),
              default=None,
              help='(lat, lon) coordinates of interest')
@click.option('--gid', '-g', type=int, default=None,
              help='Resource gid of interest')
@click.pass_context
def site(ctx, dataset, lat_lon, gid):
    """
    Extract a single dataset for the nearest pixel to the given (lat, lon)
    coordinates OR the given resource gid
    """
    if lat_lon is None and gid is None:
        click.echo("Must supply '--lat-lon' OR '--gid'!")
        raise click.Abort()
    elif lat_lon and gid:
        click.echo("You must only supply '--lat-lon' OR '--gid'!")
        raise click.Abort()

    with ctx.obj['CLS'](ctx.obj['H5'], compute_tree=ctx.obj['TREE']) as f:
        if lat_lon is not None:
            site_df = f.get_lat_lon_df(dataset, lat_lon)
        elif gid is not None:
            site_df = f.get_gid_df(dataset, gid)

    gid = site_df.name
    out_path = "{}-{}.csv".format(dataset, gid)
    out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
    logger.info('Saving data to {}'.format(out_path))
    site_df.to_csv(out_path)


@main.command()
@click.option('--dataset', '-d', type=str, required=True,
              help='Dataset to extract')
@click.option('--region', '-r', type=str, required=True,
              help='Region to extract')
@click.option('--region_col', '-col', type=str, default='state',
              help='Meta column to search for region')
@click.pass_context
def region(ctx, dataset, region, region_col):
    """
    Extract a single dataset for all pixels in the given region
    """
    with ctx.obj['CLS'](ctx.obj['H5'], compute_tree=ctx.obj['TREE']) as f:
        region_df = f.get_region_df(dataset, region, region_col=region_col)
        meta = f['meta']

    out_path = "{}-{}.csv".format(dataset, region)
    out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
    logger.info('Saving data to {}'.format(out_path))
    region_df.to_csv(out_path)

    out_path = "{}-meta.csv".format(region)
    out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
    meta = meta.loc[region_df.columns]
    meta.index.name = 'gid'
    logger.info('Saving meta data to {}'.format(out_path))
    meta.to_csv(out_path)


@main.command()
@click.option('--timestep', '-ts', type=str, required=True,
              help='Timestep to extract')
@click.option('--dataset', '-d', type=str, required=True,
              help='Dataset to extract')
@click.option('--region_col', '-col', type=str, default='state',
              help='Meta column to search for region')
@click.option('--region', '-r', type=str, default=None,
              help='Region to extract')
@click.pass_context
def timestep(ctx, timestep, dataset, region_col, region):
    """
    Extract a single dataset for a single timestep
    Extract only pixels in region if given.
    """
    with ctx.obj['CLS'](ctx.obj['H5'], compute_tree=ctx.obj['TREE']) as f:
        map_df = f.get_timestep_map(dataset, timestep, region=region,
                                    region_col=region_col)

    out_path = "{}-{}.csv".format(dataset, timestep)
    out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
    logger.info('Saving data to {}'.format(out_path))
    map_df.to_csv(out_path)


@main.group()
@click.option('--sites', '-s', type=click.Path(exists=True), required=True,
              help=('.csv or .json file with columns "latitude", "longitude" '
                    'OR "gid"'))
@click.pass_context
def multi_site(ctx, sites):
    """
    Extract multiple sites given in '--sites' .csv or .json as
    "latitude", "longitude" pairs OR "gid"s
    """
    name = os.path.splitext(os.path.basename(sites))[0]
    ctx.obj['NAME'] = name
    if sites.endswith('.csv'):
        sites = pd.read_csv(sites)
    elif sites.endswith('.json'):
        sites = pd.read_json(sites)
    else:
        click.echo("'--sites' must be a .csv or .json file!")
        click.Abort()

    if 'gid' in sites:
        ctx.obj['GID'] = sites['gid'].values
        ctx.obj['LAT_LON'] = None
    elif 'latitude' in sites and 'longitude' in sites:
        ctx.obj['GID'] = None
        ctx.obj['LAT_LON'] = sites[['latitude', 'longitude']].values
    else:
        click.echo('Must supply site "gid"s or "latitude" and "longitude" '
                   'as columns in "--sites" file')


@multi_site.command()
@click.option('--dataset', '-d', type=str, required=True,
              help='Dataset to extract')
@click.pass_context
def dataset(ctx, dataset):
    """
    Extract given dataset for all sites
    """
    gid = ctx.obj['GID']
    lat_lon = ctx.obj['LAT_LON']
    with ctx.obj['CLS'](ctx.obj['H5'], compute_tree=ctx.obj['TREE']) as f:
        meta = f['meta']
        if lat_lon is not None:
            site_df = f.get_lat_lon_df(dataset, lat_lon)
        elif gid is not None:
            site_df = f.get_gid_df(dataset, gid)

    name = ctx.obj['NAME']
    out_path = "{}-{}.csv".format(dataset, name)
    out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
    logger.info('Saving data to {}'.format(out_path))
    site_df.to_csv(out_path)

    out_path = "{}-meta.csv".format(name)
    out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
    meta = meta.loc[site_df.columns]
    meta.index.name = 'gid'
    logger.info('Saving meta data to {}'.format(out_path))
    meta.to_csv(out_path)


@multi_site.command()
@click.pass_context
def sam(ctx):
    """
    Extract SAM variables
    """
    gid = ctx.obj['GID']
    lat_lon = ctx.obj['LAT_LON']
    with ctx.obj['CLS'](ctx.obj['H5'], compute_tree=ctx.obj['TREE']) as f:
        meta = f['meta']
        if lat_lon is not None:
            SAM_df = f.get_SAM_lat_lon(lat_lon)
        elif gid is not None:
            SAM_df = f.get_SAM_gid(gid)

    name = ctx.obj['NAME']
    gids = []
    for df in SAM_df:
        gids.append(int(df.name.split('-')[-1]))
        out_path = "{}-{}.csv".format(df.name, name)
        out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
        logger.info('Saving data to {}'.format(out_path))
        df.to_csv(out_path)

    out_path = "{}-meta.csv".format(name)
    out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
    meta = meta.loc[gids]
    meta.index.name = 'gid'
    logger.info('Saving meta data to {}'.format(out_path))
    meta.to_csv(out_path)


if __name__ == '__main__':
    main(obj={})



# -*- coding: utf-8 -*-
"""
reVX-plexos utilities
"""
import copy
import json
import numpy as np
import pandas as pd
import logging
from scipy.spatial import cKDTree
from warnings import warn


logger = logging.getLogger(__name__)


def get_coord_labels(df):
    """Retrieve the coordinate labels from df.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with each row being a geo location and two columns
        containing coordinate labels.

    Returns
    -------
    df_coord_labels : list | None
        Two entry list if coordinate labels are found:
        ['lat', 'lon'] or ['latitude', 'longitude']
    """

    df_coord_labels = None
    if 'lat' in df and 'lon' in df:
        df_coord_labels = ['lat', 'lon']
    elif 'latitude' in df and 'longitude' in df:
        df_coord_labels = ['latitude', 'longitude']

    return df_coord_labels


def parse_table_name(name, wait=300, db_host='gds_edit.nrel.gov',
                     db_user=None, db_pass=None, db_port=5432):
    """Parse a dataframe from an input name.

    Parameters
    ----------
    name : str | pd.DataFrame
        CSV file path or database.schema.name or already extracted df.
    wait : int
        Integer seconds to wait for DB connection to become available
        before raising exception.
    db_host : str
        Database host name.
    db_user : str
        Your database user name.
    db_pass : str
        Database password (None if your password is cached).
    db_port : int
        Database port.

    Returns
    -------
    df : pd.DataFrame
        Extracted table
    """

    if isinstance(name, str):
        if name.endswith('.csv'):
            df = pd.read_csv(name)
        elif len(name.split('.')) == 3:
            from reVX.handlers.database import Database
            db, schema, table = name.split('.')
            logger.debug('Retrieving "{}.{}" from database "{}"'
                         .format(schema, table, db))
            df = Database.get_table(table, schema, db, wait=wait,
                                    db_host=db_host, db_user=db_user,
                                    db_pass=db_pass, db_port=db_port)

    elif isinstance(name, pd.DataFrame):
        df = name

    else:
        raise TypeError('Could not recognize input table name: '
                        '{} with type {}'.format(name, type(name)))

    return df


class DataCleaner:
    """Class for custom Plexos data cleaning procedures."""

    # Keys are bad values, values are corrected values

    REEDS_NAME_MAP = {'capacity_reV': 'built_capacity',
                      'capacity_rev': 'built_capacity',
                      'year': 'reeds_year',
                      'Year': 'reeds_year'}

    REV_NAME_MAP = {'sq_km': 'area_sq_km',
                    'capacity': 'potential_capacity',
                    'resource_ids': 'res_gids',
                    'resource_ids_cnts': 'gid_counts'}

    PLEXOS_META_COLS = ('gid', 'plexos_id', 'latitude', 'longitude',
                        'voltage', 'interconnect', 'built_capacity')

    def __init__(self, plexos_meta, profiles):
        """
        Parameters
        ----------
        plexos_meta : pd.DataFrame
            Plexos meta data including the built capacity at each plexos node.
        profiles : np.ndarray
            2D timeseries array of generation profiles. Number of columns must
            match the length of the meta data.
        """
        if profiles.shape[1] != len(plexos_meta):
            raise ValueError('Plexos profiles shape does not match meta.')

        self._plexos_meta = plexos_meta
        self._profiles = profiles

    @staticmethod
    def rename_cols(df, name_map):
        """
        Parameters
        ----------
        df : pd.DataFrame
            Input df with bad or inconsistent column names.

        Parameters
        ----------
        df : pd.DataFrame
            Same as inputs but with better col names.
        """
        df = df.rename(columns=name_map)
        return df

    @staticmethod
    def reduce_df(df, cols):
        """Reduce a df to just certain columns.

        Parameters
        ----------
        df : pd.DataFrame
            Dataframe to reduce.
        cols : list | tuple
            List of column names to keep.

        Returns
        -------
        df : pd.DataFrame
            Dataframe with only cols if the input df had all cols.
        """

        cols = [c for c in cols if c in df]
        return df[cols]

    @staticmethod
    def pre_filter_plexos_meta(plexos_meta):
        """Pre-filter the plexos meta data to drop bad node names and
        duplicate lat/lons.

        Parameters
        ----------
        meta_final : pd.DataFrame
            Plexos meta data.

        Returns
        -------
        meta_final : pd.DataFrame
            Filtered plexos meta data.
        """

        # as of 8/2019 there were two erroneous plexos nodes with bad names
        mask = (plexos_meta['plexos_id'] != '#NAME?')
        plexos_meta = plexos_meta[mask]

        # Several plexos nodes share the same location. As of 8/2019
        # Josh Novacheck suggests that the duplicate locations can be dropped.
        plexos_meta = plexos_meta.sort_values(by='voltage', ascending=False)
        plexos_meta = plexos_meta.drop_duplicates(
            subset=['latitude', 'longitude'], keep='first')
        plexos_meta = plexos_meta.sort_values(by='gid')

        return plexos_meta

    @staticmethod
    def _merge_plexos_meta(meta_final, meta_orig, i_final, i_orig):
        """Ammend the plexos meta dataframe with data about resource buildouts.

        Parameters
        ----------
        meta_final : pd.DataFrame
            Plexos meta data for the final set of nodes.
        meta_orig : pd.DataFrame
            Plexos meta data for the original pre-merge set of nodes.
        i_final : int
            Index location (iloc) of the persistent meta data row in
            meta_final.
        i_orig : int
            Index location (iloc) of the meta data row to be merged in
            meta_orig.

        Returns
        -------
        meta_final : pd.DataFrame
            Plexos meta data for the final set of nodes.
        """

        i_final = meta_final.index.values[i_final]
        i_orig = meta_orig.index.values[i_orig]

        cols = ['res_gids', 'gen_gids', 'res_built', 'built_capacity']

        for col in cols:
            val_final = meta_final.loc[i_final, col]
            val_orig = meta_orig.loc[i_orig, col]

            if not isinstance(val_final, type(val_orig)):
                raise TypeError('Mismatch in column dtype for plexos meta!')

            if isinstance(val_final, str):
                val_final = json.loads(val_final)
                val_orig = json.loads(val_orig)
                val_final += val_orig
                val_final = str(val_final)
            else:
                val_final += val_orig

            meta_final.loc[i_final, col] = val_final

        return meta_final

    def merge_small(self, capacity_threshold=20.0):
        """Merge small plexos buildout nodes into closest bigger nodes.

        Parameters
        ----------
        capacity_threshold : float
            Capacity threshold, nodes with built capacities less than this
            will be merged into bigger nodes.

        Returns
        -------
        meta : pd.DataFrame
            New plexos node meta data with updated built capacities.
        profiles : np.ndarray
            New profiles with big nodes having absorbed additional generation
            from bigger nodes.
        """

        small = (self._plexos_meta['built_capacity'] < capacity_threshold)
        big = (self._plexos_meta['built_capacity'] >= capacity_threshold)

        n_nodes = np.sum(big)
        if (n_nodes == len(self._plexos_meta) or n_nodes == 0):
            meta = None
            profiles = None

        else:
            meta = self._plexos_meta[big]
            profiles = self._profiles[:, big.values]
            logger.info('Merging plexos nodes from {} to {} due to small '
                        'nodes.'.format(len(self._plexos_meta), len(meta)))

            labels = get_coord_labels(self._plexos_meta)
            tree = cKDTree(meta[labels])
            _, nn_ind = tree.query(self._plexos_meta[labels], k=len(meta))

            for i in range(len(self._plexos_meta)):
                if small.values[i]:
                    for nn in nn_ind[i, :]:
                        if big.values[nn]:
                            meta = self._merge_plexos_meta(meta,
                                                           self._plexos_meta,
                                                           nn, i)
                            profiles[:, nn] += self._profiles[:, i]

                            break

        return meta, profiles

    def merge_extent(self, new_meta, new_profiles):
        """Merge a new set of plexos node aggregation data into the self attr.

        Parameters
        ----------
        new_meta : pd.DataFrame
            A new set of Plexos node meta data to be merged into the meta in
            self.
        new_profiles : np.ndarray
            A new set of plexos node profiles corresponding to new_meta to be
            merged into the profiles in self where the meta data overlaps with
            common nodes.
        """

        keep_index = []

        logger.info('Merging extents with {} and {} nodes ({} total).'
                    .format(len(self._plexos_meta), len(new_meta),
                            len(self._plexos_meta) + len(new_meta)))

        for i, ind in enumerate(new_meta.index.values):
            lookup = (self._plexos_meta.gid.values == new_meta.loc[ind, 'gid'])
            if any(lookup):
                i_self = np.where(lookup)[0]
                if len(i_self) > 1:
                    warn('Duplicate PLEXOS node GIDs in base plexos meta!')
                else:
                    i_self = i_self[0]

                logger.debug('Merging plexos node IDs {} and {} '
                             '(gids {} and {})'.format(
                                 self._plexos_meta.iloc[i_self]['plexos_id'],
                                 new_meta.iloc[i]['plexos_id'],
                                 self._plexos_meta.iloc[i_self]['gid'],
                                 new_meta.iloc[i]['gid']))

                self._merge_plexos_meta(self._plexos_meta, new_meta, i_self, i)
                self._profiles[:, i_self] += new_profiles[:, i]
            else:
                keep_index.append(i)

        new_meta = new_meta.loc[new_meta.index.values[keep_index]]
        new_profiles = new_profiles[:, keep_index]

        self._plexos_meta = pd.concat([self._plexos_meta, new_meta], axis=0,
                                      ignore_index=True)
        self._profiles = np.hstack((self._profiles, new_profiles))

        logger.info('Merged extents. Output has {} nodes.'
                    .format(len(self._plexos_meta)))

    def merge_multiple_extents(self, meta_list, profile_list):
        """Merge multiple plexos extents into the self attrs.

        Parameters
        ----------
        meta_list : list
            List of new meta data extents to merge into self.
        profile_list : list
            List of new gen profile to merge into self.

        Returns
        -------
        meta : pd.DataFrame
            Merged plexos node meta data.
        profiles : np.ndarray
            New profiles with merged profiles for matching nodes.
        """

        for i, meta in enumerate(meta_list):
            self.merge_extent(meta, profile_list[i])

        return self._plexos_meta, self._profiles


class ProjectGidHandler:
    """Class to handle project GIDs for a plexos project.
    Can be used to make gid superset project points for 5min data."""

    @staticmethod
    def get_resource_gids(sc_table, reeds_build, wait=300,
                          db_host='gds_edit.nrel.gov',
                          db_user=None, db_pass=None, db_port=5432):
        """Get resource gids from a  single reeds supply curve build

        Parameters
        ----------
        sc_table : str | pd.DataFrame
            reV supply curve results (CSV file path or database.schema.name)
        reeds_build : str | pd.DataFrame
            REEDS buildout file with
        wait : int
            Integer seconds to wait for DB connection to become available
            before raising exception.
        db_host : str
            Database host name.
        db_user : str
            Your database user name.
        db_pass : str
            Database password (None if your password is cached).
        db_port : int
            Database port.

        Returns
        -------
        gids : list
            Sorted list of unique integer resource gids build out.
        """

        sc_table = parse_table_name(sc_table, wait=wait,
                                    db_host=db_host,
                                    db_user=db_user,
                                    db_pass=db_pass,
                                    db_port=db_port)
        reeds_build = parse_table_name(reeds_build, wait=wait,
                                       db_host=db_host,
                                       db_user=db_user,
                                       db_pass=db_pass,
                                       db_port=db_port)

        rev_name_map = copy.deepcopy(DataCleaner.REV_NAME_MAP)
        reeds_name_map = copy.deepcopy(DataCleaner.REEDS_NAME_MAP)
        rev_name_map['sc_gids'] = 'gid'
        rev_name_map['sc_gid'] = 'gid'
        rev_name_map['gids'] = 'gid'
        reeds_name_map['sc_gids'] = 'gid'
        reeds_name_map['gids'] = 'gid'
        sc_table = DataCleaner.rename_cols(sc_table, rev_name_map)
        reeds_build = DataCleaner.rename_cols(reeds_build, reeds_name_map)

        reeds_gids = reeds_build['gid'].values.tolist()
        rev_gids = sc_table['gid'].values.tolist()

        missing = [gid for gid in reeds_gids if gid not in rev_gids]
        if any(missing):
            e = ('The following gids were built in reeds but not found in '
                 'the reV sc table: {}'.format(missing))
            logger.error(e)
            raise RuntimeError(e)

        gid_table = pd.merge(reeds_build, sc_table, how='left', on='gid')

        gids = []
        for res_gid_list in gid_table['res_gids'].values.tolist():
            if isinstance(res_gid_list, str):
                res_gid_list = json.loads(res_gid_list)
            gids += [int(gid) for gid in res_gid_list]

        if not any(gids):
            e = 'No resource gids found!'
            logger.error(e)
            raise ValueError(e)

        gids = sorted(list(set(gids)), key=float)

        return gids

    @staticmethod
    def build_project_points(build_map, fpath_out=None, config_tag='default',
                             **db_kwargs):
        """Build a project points CSV from a set of rev/reeds build files.

        Parameters
        ----------
        build_map : dict
            Mapping of buildout files/tables. Keys are filepaths to reeds
            buildout files, values are reV SC tables (can be db names).
        fpath_out : str | None
            Output filepath to save project points file.
        config_tag : str
            Config tab/label to write to the project points config column.
        db_kwargs : dict
            Optional database kwargs.

        Returns
        -------
        pp : pd.DataFrame
            Project points dataframe with gid and config columns..
        """
        tables = {}
        gids = []

        for reeds_table, rev_table in build_map.items():
            if reeds_table not in tables:
                tables[reeds_table] = parse_table_name(reeds_table,
                                                       **db_kwargs)
            if rev_table not in tables:
                tables[rev_table] = parse_table_name(rev_table, **db_kwargs)

            gids += ProjectGidHandler.get_resource_gids(tables[rev_table],
                                                        tables[reeds_table])

        gids = sorted(list(set(gids)), key=float)
        pp = pd.DataFrame({'config': [config_tag] * len(gids)}, index=gids)
        pp.index.name = 'gid'

        if fpath_out:
            logger.debug('Writing project points: {}'.format(fpath_out))
            pp.to_csv(fpath_out)

        return pp



# -*- coding: utf-8 -*-
"""
Resource Extraction Tools
"""
import gzip
import logging
import numpy as np
import os
import pandas as pd
import pickle
from reV.handlers.resource import Resource
from reV.handlers.rev_resource import (MultiFileWTK, MultiFileNSRDB, NSRDB,
                                       SolarResource, WindResource)
from scipy.spatial import cKDTree

BIN = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
BIN = os.path.join(os.path.dirname(BIN), 'bin')
logger = logging.getLogger(__name__)


class ResourceX(Resource):
    """
    Resource data extraction tool
    """
    def __init__(self, res_h5, tree=None, unscale=True, hsds=False,
                 str_decode=True, group=None):
        """
        Parameters
        ----------
        res_h5 : str
            Path to resource .h5 file of interest
        tree : str
            Path to .pgz file containing pickled cKDTree of lat, lon
            coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        group : str
            Group within .h5 resource file to open
        """
        super().__init__(res_h5, unscale=unscale, hsds=hsds,
                         str_decode=str_decode, group=group)
        self._tree = tree

    @property
    def tree(self):
        """
        Returns
        -------
        tree : cKDTree
            Lat, lon coordinates cKDTree
        """
        if not isinstance(self._tree, cKDTree):
            self._tree = self._init_tree(tree=self._tree)

        return self._tree

    @property
    def countries(self):
        """
        Returns
        -------
        countries : ndarray
            Countries available in .h5 file
        """
        if 'country' in self.meta:
            countries = self.meta['country'].unique()
        else:
            countries = None

        return countries

    @property
    def states(self):
        """
        Returns
        -------
        states : ndarray
            States available in .h5 file
        """
        if 'state' in self.meta:
            states = self.meta['state'].unique()
        else:
            states = None

        return states

    @property
    def counties(self):
        """
        Returns
        -------
        counties : ndarray
            Counties available in .h5 file
        """
        if 'county' in self.meta:
            counties = self.meta['county'].unique()
        else:
            counties = None

        return counties

    @staticmethod
    def _load_tree(tree_pickle):
        """
        Load tree from pickle file

        Parameters
        ----------
        tree_pickle : str
            Pickle (.pkl, .pickle) or compressed pickle (.pgz, .pgzip) file
            containing precomputed cKDTree

        Returns
        -------
        tree : cKDTree
            Precomputed tree of lat, lon coordinates
        """
        try:
            if tree_pickle.endswith(('.pkl', '.pickle')):
                with open(tree_pickle, 'rb') as f:
                    tree = pickle.load(f)
            elif tree_pickle.endswith(('.pgz', '.pgzip', '.gz', '.gzip')):
                with gzip.open(tree_pickle, 'r') as f:
                    tree = pickle.load(f)
            else:
                logger.warning('Cannot parse files of type "{}"'
                               .format(tree_pickle))
                tree = None
        except Exception as e:
            logger.warning('Could not extract tree from {}: {}'
                           .format(tree_pickle, e))
            tree = None

        return tree

    def _init_tree(self, tree=None, compute_tree=False):
        """
        Inititialize cKDTree of lat, lon coordinates

        Parameters
        ----------
        tree : str | cKDTree | NoneType
            Path to .pgz file containing pre-computed tree
            If None search bin for .pgz file matching h5 file
            else compute tree
        compute_tree : bool
            Force the computation of the cKDTree

        Returns
        -------
        tree : cKDTree
            cKDTree of lat, lon coordinate from wtk .h5 file
        """
        if compute_tree:
            tree = None
        else:
            if not isinstance(tree, (cKDTree, str, type(None))):
                tree = None
                logger.warning('Precomputed tree must be supplied as a pickle '
                               'file or a cKDTree, not a {}'
                               .format(type(tree)))

            if tree is None:
                pgz_files = [file for file in os.listdir(BIN)
                             if file.endswith('.pgz')]
                for pgz in pgz_files:
                    prefix = pgz.split('_tree')[0]
                    if self.h5_file.startswith(prefix):
                        tree = os.path.join(BIN, pgz)
                        break

            if isinstance(tree, str):
                tree = self._load_tree(tree)

        if tree is None:
            lat_lon = self.meta[['latitude', 'longitude']].values
            tree = cKDTree(lat_lon)

        return tree

    def _get_nearest(self, lat_lon):
        """
        Get nearest gid to given (lat, lon) pair or pairs

        Parameters
        ----------
        lat_lon : ndarray
            Either a single (lat, lon) pair or series of (lat, lon) pairs

        Returns
        -------
        gids : int | ndarray
            Nearest gid(s) to given (lat, lon) pair(s)
        """
        _, gids = self.tree.query(lat_lon)
        return gids

    def _get_region(self, region, region_col='state'):
        """
        Get the gids for given region

        Parameters
        ----------
        region : str
            Region to search for
        region_col : str
            Region column to search

        Returns
        -------
        gids : ndarray
            Vector of gids in given region
        """
        gids = (self.meta[region_col] == region).index.values
        return gids

    def _get_timestep_idx(self, timestep):
        """
        Get the index of the desired timestep

        Parameters
        ----------
        timestep : str
            Timestep of interest

        Returns
        -------
        ts_idx : int
            Time index value
        """
        timestep = pd.to_datetime(timestep)
        idx = np.where(self.time_index == timestep)[0][0]

        return idx

    def get_gid_ts(self, ds_name, gid):
        """
        Extract timeseries of site(s) neareset to given lat_lon(s)

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        gid : int | list
            Resource gid(s) of interset

        Return
        ------
        site_ts : ndarray
            Time-series for given site(s) and dataset
        """
        site_ts = self[ds_name, :, gid]

        return site_ts

    def get_gid_df(self, ds_name, gid):
        """
        Extract timeseries of site(s) nearest to given lat_lon(s) and return
        as a DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        gid : int | list
            Resource gid(s) of interset

        Return
        ------
        site_df : pandas.DataFrame
            Time-series DataFrame for given site and dataset
        """
        if isinstance(gid, int):
            site_df = pd.DataFrame({ds_name: self[ds_name, :, gid]},
                                   index=self.time_index)
            site_df.name = gid
            site_df.index.name = 'time_index'
        else:
            site_df = pd.DataFrame(self[ds_name, :, gid], columns=gid,
                                   index=self.time_index)
            site_df.name = ds_name
            site_df.index.name = 'time_index'

        return site_df

    def get_lat_lon_ts(self, ds_name, lat_lon):
        """
        Extract timeseries of site(s) neareset to given lat_lon(s)

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        lat_lon : tuple | list
            (lat, lon) coordinate of interest or pairs of coordinates

        Return
        ------
        site_ts : ndarray
            Time-series for given site(s) and dataset
        """
        gid = self._get_nearest(lat_lon)
        site_ts = self.get_gid_ts(ds_name, gid)

        return site_ts

    def get_lat_lon_df(self, ds_name, lat_lon):
        """
        Extract timeseries of site(s) nearest to given lat_lon(s) and return
        as a DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        lat_lon : tuple
            (lat, lon) coordinate of interest

        Return
        ------
        site_df : pandas.DataFrame
            Time-series DataFrame for given site and dataset
        """
        gid = self._get_nearest(lat_lon)
        site_df = self.get_gid_df(ds_name, gid)

        return site_df

    def get_region_ts(self, ds_name, region, region_col='state'):
        """
        Extract timeseries of of all sites in given region

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        region : str
            Region to search for
        region_col : str
            Region column to search

        Return
        ------
        region_ts : ndarray
            Time-series array of desired dataset for all sites in desired
            region
        """
        gids = self._get_region(region, region_col=region_col)
        region_ts = self[ds_name, :, gids]

        return region_ts

    def get_region_df(self, ds_name, region, region_col='state'):
        """
        Extract timeseries of of all sites in given region and return as a
        DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        region : str
            Region to extract all pixels for
        region_col : str
            Region column to search

        Return
        ------
        region_df : pandas.DataFrame
            Time-series array of desired dataset for all sites in desired
            region
        """
        gids = self._get_region(region, region_col=region_col)
        region_df = pd.DataFrame(self[ds_name, :, gids], columns=gids,
                                 index=self.time_index)
        region_df.name = ds_name
        region_df.index.name = 'time_index'

        return region_df

    def get_SAM_gid(self, gid):
        """
        Extract time-series of all variables needed to run SAM for nearest
        site to given resource gid

        Parameters
        ----------
        gid : int | list
            Resource gid(s) of interset

        Return
        ------
        SAM_df : pandas.DataFrame | list
            Time-series DataFrame for given site and dataset
            If multiple lat, lon pairs are given a list of DatFrames is
            returned
        """
        if isinstance(gid, int):
            gid = [gid, ]

        SAM_df = []
        for id in gid:
            SAM_df.append(self['SAM', id])

        if len(SAM_df) == 1:
            SAM_df = SAM_df[0]

        return SAM_df

    def get_SAM_lat_lon(self, lat_lon):
        """
        Extract time-series of all variables needed to run SAM for nearest
        site to given lat_lon

        Parameters
        ----------
        lat_lon : tuple
            (lat, lon) coordinate of interest

        Return
        ------
        SAM_df : pandas.DataFrame | list
            Time-series DataFrame for given site and dataset
            If multiple lat, lon pairs are given a list of DatFrames is
            returned
        """
        gid = self._get_nearest(lat_lon)
        SAM_df = self.get_SAM_gid(gid)

        return SAM_df

    def get_timestep_map(self, ds_name, timestep, region=None,
                         region_col='state'):
        """
        Extract a map of the given dataset at the given timestep for the
        given region if supplied

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        timestep : str
            Timestep of interest
        region : str
            Region to extract all pixels for
        region_col : str
            Region column to search

        Returns
        -------
        ts_map : pandas.DataFrame
            DataFrame of map values
        """
        lat_lons = self.meta[['latitude', 'longitude']].values
        ts_idx = self._get_timestep_idx(timestep)
        gids = slice(None)
        if region is not None:
            gids = self._get_region(region, region_col=region_col)
            lat_lons = lat_lons[gids]

        ts_map = self[ds_name, ts_idx, gids]
        ts_map = pd.DataFrame({'longitude': lat_lons[:, 1],
                               'latitude': lat_lons[:, 0],
                               ds_name: ts_map})

        return ts_map


class SolarX(SolarResource, ResourceX):
    """
    Solar Resource extraction class
    """
    def __init__(self, solar_h5, tree=None, compute_tree=False, unscale=True,
                 hsds=False, str_decode=True, group=None):
        """
        Parameters
        ----------
        solar_h5 : str
            Path to solar .h5 file of interest
        tree : str
            path to .pgz file containing pickled cKDTree of lat, lon
            coordinates
        compute_tree : bool
            Force the computation of the cKDTree
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        group : str
            Group within .h5 resource file to open
        """
        super().__init__(solar_h5, unscale=unscale, hsds=hsds,
                         str_decode=str_decode, group=group)
        self._tree = self._init_tree(tree=tree, compute_tree=compute_tree)


class NSRDBX(NSRDB, ResourceX):
    """
    NSRDB extraction class
    """
    def __init__(self, nsrdb_h5, tree=None, compute_tree=False, unscale=True,
                 hsds=False, str_decode=True, group=None):
        """
        Parameters
        ----------
        nsrdb_h5 : str
            Path to NSRDB .h5 file of interest
        tree : str
            path to .pgz file containing pickled cKDTree of lat, lon
            coordinates
        compute_tree : bool
            Force the computation of the cKDTree
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        group : str
            Group within .h5 resource file to open
        """
        super().__init__(nsrdb_h5, unscale=unscale, hsds=hsds,
                         str_decode=str_decode, group=group)
        self._tree = self._init_tree(tree=tree, compute_tree=compute_tree)


class MultiFileNSRDBX(MultiFileNSRDB, ResourceX):
    """
    NSRDB extraction class
    """
    def __init__(self, nsrdb_path, tree=None, compute_tree=False,
                 unscale=True, str_decode=True):
        """
        Parameters
        ----------
        nsrdb_path : str
            Path to NSRDB .h5 files
            Available formats:
                /h5_dir/
                /h5_dir/prefix*suffix
        tree : str
            path to .pgz file containing pickled cKDTree of lat, lon
            coordinates
        compute_tree : bool
            Force the computation of the cKDTree
        prefix : str
            Prefix for resource .h5 files
        suffix : str
            Suffix for resource .h5 files
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        """
        super().__init__(nsrdb_path, unscale=unscale, str_decode=str_decode)
        self._tree = self._init_tree(tree=tree, compute_tree=compute_tree)


class WindX(WindResource, ResourceX):
    """
    Wind Resource extraction class
    """
    def __init__(self, wind_h5, tree=None, compute_tree=False, unscale=True,
                 hsds=False, str_decode=True, group=None):
        """
        Parameters
        ----------
        wind_h5 : str
            Path to Wind .h5 file of interest
        tree : str
            path to .pgz file containing pickled cKDTree of lat, lon
            coordinates
        compute_tree : bool
            Force the computation of the cKDTree
        kwargs : dict
            Kwargs for Resource
        """
        super().__init__(wind_h5, unscale=unscale, hsds=hsds,
                         str_decode=str_decode, group=group)
        self._tree = self._init_tree(tree=tree, compute_tree=compute_tree)

    def get_SAM_gid(self, hub_height, gid, **kwargs):
        """
        Extract time-series of all variables needed to run SAM for nearest
        site to given resource gid and hub height

        Parameters
        ----------
        hub_height : int
            Hub height of interest
        gid : int | list
            Resource gid(s) of interset
        kwargs : dict
            Internal kwargs for _get_SAM_df:
            - require_wind_dir
            - icing

        Return
        ------
        SAM_df : pandas.DataFrame | list
            Time-series DataFrame for given site and dataset
            If multiple lat, lon pairs are given a list of DatFrames is
            returned
        """
        ds_name = 'SAM_{}m'.format(hub_height)
        if isinstance(gid, int):
            gid = [gid, ]

        SAM_df = []
        for id in gid:
            SAM_df.append(self._get_SAM_df(ds_name, id, **kwargs))

        if len(SAM_df) == 1:
            SAM_df = SAM_df[0]

        return SAM_df

    def get_SAM_lat_lon(self, hub_height, lat_lon):
        """
        Extract time-series of all variables needed to run SAM for nearest
        site to given lat_lon and hub height

        Parameters
        ----------
        hub_height : int
            Hub height of interest
        gid : int | list
            Resource gid(s) of interset
        kwargs : dict
            Internal kwargs for _get_SAM_df:
            - require_wind_dir
            - icing

        Return
        ------
        SAM_df : pandas.DataFrame | list
            Time-series DataFrame for given site and dataset
            If multiple lat, lon pairs are given a list of DatFrames is
            returned
        """
        gid = self._get_nearest(lat_lon)
        SAM_df = self.get_SAM_gid(hub_height, gid)

        return SAM_df


class MultiFileWindX(MultiFileWTK, WindX):
    """
    Wind Resource extraction class
    """
    def __init__(self, wtk_path, tree=None, compute_tree=False,
                 unscale=True, str_decode=True):
        """
        Parameters
        ----------
        wtk_path : str
            Path to five minute WTK .h5 files
            Available formats:
                /h5_dir/
                /h5_dir/prefix*suffix
        tree : str
            path to .pgz file containing pickled cKDTree of lat, lon
            coordinates
        compute_tree : bool
            Force the computation of the cKDTree
        prefix : str
            Prefix for resource .h5 files
        suffix : str
            Suffix for resource .h5 files
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        """
        super().__init__(wtk_path, unscale=unscale, str_decode=str_decode)
        self._tree = self._init_tree(tree=tree, compute_tree=compute_tree)



# -*- coding: utf-8 -*-
"""
ReEDS Command Line Interface
"""
import click
import logging
import os

from reV.utilities.loggers import init_mult
from reV.utilities.cli_dtypes import STR, STRLIST, INT
from reV.utilities.execution import SLURM, SubprocessManager
from reV.utilities.utilities import dict_str_load

from reVX.config.reeds import ReedsConfig
from reVX.reeds.reeds_classification import ReedsClassifier
from reVX.reeds.reeds_profiles import ReedsProfiles
from reVX.reeds.reeds_timeslices import ReedsTimeslices
from reVX.utilities.exceptions import ReedsRuntimeError


logger = logging.getLogger(__name__)


@click.group()
@click.option('--name', '-n', default='ReEDS', type=STR,
              help='Job name. Default is "ReEDS".')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """
    ReEDS Command Line Interface
    """
    ctx.ensure_object(dict)
    ctx.obj['NAME'] = name
    ctx.obj['VERBOSE'] = verbose


def run_local(ctx, config):
    """
    Run reV to ReEDs locally from config

    Parameters
    ----------
    ctx : click.ctx
        click ctx object
    config : ReedsConfig
        Reeds Config object
    """
    ctx.obj['NAME'] = config.name
    ctx.invoke(local,
               out_dir=config.dirout,
               log_dir=config.logdir)

    if config.classify is not None:
        ctx.invoke(classify,
                   rev_table=config.classify.rev_table,
                   resource_classes=config.classify.resource_classes,
                   regions=config.classify.regions,
                   cap_bins=config.classify.cap_bins,
                   sort_bins_by=config.classify.sort_bins_by,
                   pre_filter=config.classify.pre_filter)

    if config.profiles is not None:
        ctx.invoke(profiles,
                   reeds_table=config.profiles.reeds_table,
                   cf_profiles=config.profiles.cf_profiles,
                   gid_col=config.profiles.gid_col,
                   n_profiles=config.profiles.n_profiles,
                   profiles_dset=config.profiles.profiles_dset,
                   rep_method=config.profiles.rep_method,
                   err_method=config.profiles.err_method,
                   weight=config.profiles.weight,
                   reg_cols=config.profiles.reg_cols,
                   max_workers=config.profiles.max_workers)

    if config.timeslices is not None:
        ctx.invoke(timeslices,
                   profiles=config.timeslices.profiles,
                   timeslices=config.timeslices.timeslices,
                   reg_cols=config.timeslices.reg_cols,
                   all_profiles=config.timeslices.all_profiles)


@main.command()
@click.option('--config', '-c', required=True,
              type=click.Path(exists=True),
              help='Filepath to reVX-REEDS config json file.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config, verbose):
    """
    Run reVX-REEDS from a config.
    """

    config = ReedsConfig(config)

    if 'VERBOSE' in ctx.obj:
        if any((ctx.obj['VERBOSE'], verbose)):
            config._log_level = logging.DEBUG
    elif verbose:
        config._log_level = logging.DEBUG

    if config.execution_control.option == 'local':
        run_local(ctx, config)

    if config.execution_control.option == 'eagle':
        eagle(config)


@main.group(chain=True)
@click.option('--out_dir', '-o', required=True, type=click.Path(),
              help='Directory to dump output files')
@click.option('--log_dir', '-log', default=None, type=STR,
              help='Directory to dump log files. Default is out_dir.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def local(ctx, out_dir, log_dir, verbose):
    """
    Run reVX-REEDS on local hardware.
    """
    ctx.obj['OUT_DIR'] = out_dir
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    if log_dir is None:
        log_dir = out_dir

    name = ctx.obj['NAME']
    if 'VERBOSE' in ctx.obj:
        verbose = any((ctx.obj['VERBOSE'], verbose))

    log_modules = [__name__, 'reVX.reeds', 'reV.rep_profiles']
    init_mult(name, log_dir, modules=log_modules, verbose=verbose)

    logger.info('Running reV to ReEDS pipeline\n'
                'Outputs to be stored in: {}'.format(out_dir))


@local.command()
@click.option('--rev_table', '-rt', required=True,
              type=click.Path(exists=True),
              help=('Path to .csv containing reV aggregation or '
                    'supply curve table'))
@click.option('--resource_classes', '-rc', required=True,
              type=click.Path(exists=True),
              help=("resource_classes: str | pandas.DataFrame\n"
                    "Resource classes, either provided in a .csv, .json or a "
                    "DataFrame\n"
                    "Allowable columns:\n"
                    "- 'class' -> class labels to use\n"
                    "- 'TRG_cap' -> TRG capacity bins to use to create TRG "
                    "classes\n"
                    "- any column in 'rev_table' -> Categorical bins\n"
                    "- '*_min' and '*_max' where * is a numberical column in "
                    "'rev_table' -> Range binning\n"
                    "NOTE: 'TRG_cap' can only be combined with categorical "
                    "bins"))
@click.option('--regions', '-r', type=str, default='reeds_region',
              help='Mapping of supply curve points to geographic region')
@click.option('--cap_bins', '-cb', type=int, default=5,
              help=('Number of capacity bins to create for each '
                    'region/resource bin combination'))
@click.option('--sort_bins_by', '-sb', type=str, default='trans_cap_cost',
              help='Column(s) in rev_table to sort before binning')
@click.option('--pre_filter', '-f', type=STR, default=None,
              help='Column value pair(s) to filter on. If None do not filter')
@click.pass_context
def classify(ctx, rev_table, resource_classes, regions, cap_bins, sort_bins_by,
             pre_filter):
    """
    Extract ReEDS (region, bin, class) groups
    """
    name = ctx.obj['NAME']
    out_dir = ctx.obj['OUT_DIR']

    logger.info('Extracting ReEDS (region, bin, class) groups using '
                'reV sc table {}'.format(rev_table))
    if isinstance(pre_filter, str):
        pre_filter = dict_str_load(pre_filter)

    out = ReedsClassifier.create(rev_table, resource_classes,
                                 region_map=regions, cap_bins=cap_bins,
                                 sort_bins_by=sort_bins_by,
                                 pre_filter=pre_filter)
    table_full, table, agg_table_full, agg_table = out

    out_path = os.path.join(out_dir,
                            '{}_supply_curve_raw_full.csv'.format(name))
    table_full.to_csv(out_path, index=False)
    out_path = os.path.join(out_dir, '{}_supply_curve_raw.csv'.format(name))
    table.to_csv(out_path, index=False)
    out_path = os.path.join(out_dir, '{}_supply_curve_full.csv'.format(name))
    agg_table_full.to_csv(out_path, index=False)
    out_path = os.path.join(out_dir, '{}_supply_curve.csv'.format(name))
    agg_table.to_csv(out_path, index=False)

    ctx.obj['TABLE'] = table_full

    logger.info('reVX - ReEDS classification methods complete.')


@local.command()
@click.option('--reeds_table', '-rt', type=STR, default=None,
              help=('Path to .csv containing reeds classification table '
                    'not needed if chained with classify command'))
@click.option('--cf_profiles', '-cf', required=True,
              type=click.Path(exists=True),
              help=('Path to reV .h5 file containing desired capacity factor '
                    'profiles'))
@click.option('--gid_col', '-gc', type=str, default='gen_gids',
              help='Column label in rev_summary that contains the generation '
              'gids (data index in cf_profiles file path).')
@click.option('--n_profiles', '-np', type=int, default=1,
              help='Number of profiles to extract per "group".')
@click.option('--profiles_dset', '-pd', type=str, default="cf_profile",
              help='Profiles dataset name in cf_profiles file.')
@click.option('--rep_method', '-rm', type=STR, default='meanoid',
              help=('Method identifier for calculation of the representative '
                    'profile.'))
@click.option('--err_method', '-em', type=STR, default='rmse',
              help=('Method identifier for calculation of error from the '
                    'representative profile.'))
@click.option('--weight', '-w', type=str, default='gid_counts',
              help='Column in rev_summary used to apply weighted mean to '
              'profiles. The supply curve table data in the weight column '
              'should have weight values corresponding to the gid_col in '
              'the same row.')
@click.option('--reg_cols', '-rcp', type=STRLIST,
              default=('region', 'bin', 'class'),
              help=('Label(s) for a categorical region column(s) to extract '
                    'profiles for (default is region, bin, class)'))
@click.option('--max_workers', '-mw', type=INT, default=None,
              help=('Number of parallel workers. 1 will run serial, '
                    'None will use all available.'))
@click.pass_context
def profiles(ctx, reeds_table, cf_profiles, gid_col, n_profiles, profiles_dset,
             rep_method, err_method, weight, reg_cols, max_workers):
    """
    Extract ReEDS represntative profiles
    """
    name = ctx.obj['NAME']
    if reeds_table is None:
        reeds_table = ctx.obj['TABLE']

    out_dir = ctx.obj['OUT_DIR']

    logger.info('Extracting ReEDS representative profiles for {} groups, '
                'from {}.'
                .format(reg_cols, cf_profiles))

    out_path = os.path.join(out_dir, '{}_hourly_cf.h5'.format(name))
    logger.info('Saving representative hourly cf profiles to {}.'
                .format(out_path))

    ReedsProfiles.run(cf_profiles, reeds_table,
                      gid_col=gid_col,
                      profiles_dset=profiles_dset,
                      rep_method=rep_method,
                      err_method=err_method,
                      n_profiles=n_profiles,
                      weight=weight,
                      reg_cols=reg_cols,
                      max_workers=max_workers,
                      fout=out_path,
                      hourly=True)

    ctx.obj['PROFILES'] = out_path

    logger.info('reVX - ReEDS representative profile methods complete.')


@local.command()
@click.option('--profiles', '-pr', type=STR, default=None,
              help=('Path to .h5 file containing (representative) profiles, '
                    'not needed if chained with profiles command'))
@click.option('--timeslices', '-ts', required=True,
              type=click.Path(exists=True),
              help='.csv containing timeslice mapping')
@click.option('--reg_cols', '-rct', type=STRLIST,
              default=('region', 'class'),
              help=('Label(s) for a categorical region column(s) to create '
                    'timeslice stats for (default is region and class)'))
@click.option('--all_profiles', '-ap', is_flag=True,
              help='Flag to calculate timeslice stats from all CF profiles '
              'as opposed to representative profiles (default).')
@click.pass_context
def timeslices(ctx, profiles, timeslices, reg_cols, all_profiles):
    """
    Extract timeslices from representative profiles
    """
    name = ctx.obj['NAME']
    out_dir = ctx.obj['OUT_DIR']

    if profiles is None:
        if 'PROFILES' not in ctx.obj:
            msg = ('You must run "profiles" or provide path to existing '
                   'profiles to extract timeslices!')
            logger.error(msg)
            raise ReedsRuntimeError(msg)
        profiles = ctx.obj['PROFILES']

    logger.info('Extracting timeslices from {} using mapping {}'
                .format(profiles, timeslices))

    if all_profiles:
        rev_table = ctx.obj['rev_table']
        logger.info('Extracting timeslice stats from all cf_profiles '
                    'in rev_table (and not just representative profiles).')
    else:
        rev_table = None
        logger.info('Extracting timeslice stats from representative profiles '
                    '(and not all cf_profiles).')

    stats, corr = ReedsTimeslices.run(profiles, timeslices,
                                      rev_table=rev_table,
                                      reg_cols=reg_cols,
                                      legacy_format=False)

    out_path = os.path.join(out_dir, '{}_performance.csv'.format(name))
    logger.info('Saving timeslice performance stats to {}'.format(out_path))
    stats.to_csv(out_path, index=False)

    out_path = os.path.join(out_dir, '{}_correlations.h5'.format(name))
    logger.info('Saving timeslice correlations to {}'.format(out_path))
    ReedsTimeslices.save_correlation_dict(corr, reg_cols, out_path)

    logger.info('reVX - ReEDS timeslice methods complete.')


def get_node_cmd(config):
    """
    Get the node CLI call for the reVX-REEDS pipeline.

    Parameters
    ----------
    config : reVX.config.reeds.ReedsConfig
        reVX-REEDS config object.

    Returns
    -------
    cmd : str
        CLI call to submit to SLURM execution.
    """
    s = SubprocessManager.s

    args = ('-n {name} local -o {out_dir} -log {log_dir} '
            .format(name=s(config.name),
                    out_dir=s(config.dirout),
                    log_dir=s(config.logdir)))

    if config.log_level == logging.DEBUG:
        args += '-v '

    if config.classify is not None:
        args += ('classify -rt {rev_table} -rc {resource_classes} '
                 '-r {regions} -cb {cap_bins} -sb {sort_bins_by} '
                 '-f {pre_filter} '
                 .format(rev_table=s(config.classify.rev_table),
                         resource_classes=s(config.classify.resource_classes),
                         regions=s(config.classify.regions),
                         cap_bins=s(config.classify.cap_bins),
                         sort_bins_by=s(config.classify.sort_bins_by),
                         pre_filter=s(config.classify.pre_filter)))

    if config.profiles is not None:
        args += ('profiles -rt {reeds_table} -cf {cf_profiles} -gc {gid_col} '
                 '-np {n_profiles} -pd {profiles_dset} -rm {rep_method} '
                 '-em {err_method} -w {weight} -rcp {reg_cols} '
                 '-mw {max_workers} '
                 .format(reeds_table=s(config.profiles.reeds_table),
                         cf_profiles=s(config.profiles.cf_profiles),
                         gid_col=s(config.profiles.gid_col),
                         n_profiles=s(config.profiles.n_profiles),
                         profiles_dset=s(config.profiles.profiles_dset),
                         rep_method=s(config.profiles.rep_method),
                         err_method=s(config.profiles.err_method),
                         weight=s(config.profiles.weight),
                         reg_cols=s(config.profiles.reg_cols),
                         max_workers=s(config.profiles.max_workers)))

    if config.timeslices is not None:
        args += ('timeslices -pr {profiles} -ts {timeslices} -rct {reg_cols} '
                 .format(profiles=s(config.timeslices.profiles),
                         timeslices=s(config.timeslices.timeslices),
                         reg_cols=s(config.timeslices.reg_cols)))
        if config.timeslices.all_profiles:
            args += '-ap '

    cmd = 'python -m reVX.reeds.reeds_cli {}'.format(args)
    logger.debug('Submitting the following cli call:\n\t{}'.format(cmd))
    return cmd


def eagle(config):
    """
    Run reVX-REEDS on Eagle HPC.

    Parameters
    ----------
    config : reVX.config.reeds.ReedsConfig
        reVX-REEDS config object.
    """

    cmd = get_node_cmd(config)
    name = config.name
    log_dir = config.logdir
    stdout_path = os.path.join(log_dir, 'stdout/')

    logger.info('Running reVX-REEDS pipeline on Eagle with '
                'node name "{}"'.format(name))
    slurm = SLURM(cmd, alloc=config.execution_control.alloc,
                  memory=config.execution_control.node_mem,
                  walltime=config.execution_control.walltime,
                  feature=config.execution_control.feature,
                  name=name, stdout_path=stdout_path,
                  conda_env=config.execution_control.conda_env,
                  module=config.execution_control.module)
    if slurm.id:
        msg = ('Kicked off reVX-REEDS pipeline job "{}" '
               '(SLURM jobid #{}) on Eagle.'
               .format(name, slurm.id))
    else:
        msg = ('Was unable to kick off reVX-REEDS pipeline job "{}". '
               'Please see the stdout error messages'
               .format(name))

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running ReEDS CLI')
        raise
